import math
from collections import Counter
from typing import Dict, Callable, List

import numpy as np
import timeit

import pandas as pd
from sklearn import metrics as M


def clustering_report(y_true, y_pred) -> pd.DataFrame:
    """
    Generate cluster evaluation metrics.

    Args:
        y_true: Array of actual labels
        y_pred: Array of predicted clusters

    Returns:
        Pandas DataFrame with metrics.
    """
    return pd.DataFrame({
        'Homogeneity': M.homogeneity_score(y_true, y_pred),
        'Completeness': M.completeness_score(y_true, y_pred),
        'V-Measure': M.v_measure_score(y_true, y_pred),
        'Adjusted Rand Index': M.adjusted_rand_score(y_true, y_pred),
        'Adjusted Mutual Information': M.adjusted_mutual_info_score(y_true, y_pred)
    }, index=['value']).T


def multilabel_classification_report(y_true, y_pred) -> pd.Series:
    """
    Compute all metrics for a multi-label classification problem.
    Args:
        y_true: True binarized labels
        y_pred: Predicted binarized labels

    Returns:
        Pandas series of metrics
    """
    scores = {
        'accuracy': M.accuracy_score(y_true, y_pred),
        'precision_macro': M.precision_score(y_true, y_pred, average='macro'),
        'recall_macro': M.recall_score(y_true, y_pred, average='macro'),
        'f1_macro': M.f1_score(y_true, y_pred, average='macro'),
        'f1_weighted': M.f1_score(y_true, y_pred, average='weighted'),
        'hamming_loss': M.hamming_loss(y_true, y_pred)
    }
    return pd.Series(scores)


def benchmark_function(fn: Callable,
                       repeat: int = 5) -> Dict:
    """
    Benchmark time taken for a function and return metrics.

    Args:
        fn: A python function
        repeat: Number of samples

    Returns:
        Dictionary of total times, mean and std of times
    """
    iteration_times = timeit.repeat(fn,
                                    repeat=repeat,
                                    number=1)
    return {'time': iteration_times,
            'mean': np.mean(iteration_times),
            'std': np.std(iteration_times)}


def baseline_accuracy(labels: List) -> float:
    """
    Get accuracy for always majority class classifier.

    Usage:
    ```python
    >>> baseline_accuracy([0, 1])
    50.0
    ```

    Args:
        labels: List of class labels.

    Returns:
        Baseline accuracy
    """
    (label, count), *_ = Counter(labels).most_common(1)
    return count / len(labels) * 100.0


def missing_value_percent(df):
    """
    Get the percentage of missing values in each column.

    Args:
        df: Pandas DataFrame

    Returns:
        Percentage of missing value in each column.
    """
    num_rows = len(df)
    return (df.isna().sum() / num_rows * 100.0).sort_values(ascending=False)


def na_percent(df):
    return missing_value_percent(df)


def n_clusters(data) -> int:
    """
    Generate number of clusters to create.

    Heuristic:
    Number of clusters = square root of total data points

    Args:
        data: Total number of data points or the data point itself

    Returns:
        Number of clusters
    """
    if type(data) is int:
        total_rows = data
    else:
        total_rows = len(set(data))
    return int(math.sqrt(total_rows))


def vector_similarity(u, v) -> float:
    angle = np.dot(u, v) / (np.linalg.norm(u) * np.linalg.norm(v))
    return float(angle)


def jaccard(x, y) -> float:
    """
    Compute jaccard similarity (intersection over union).

    Args:
        x: Array-like object
        y: Array-like object

    Returns:
        Intersection Over Union score
    """
    s1 = set(x)
    s2 = set(y)
    if len(s1) == 0 and len(s2) == 0:
        return 0
    return len(s1 & s2) / len(s1 | s2)


def sorted_classification_report(y_true, y_pred, **kwargs) -> pd.DataFrame:
    """
    Generate class-wise classification report sorted from worst to best.

    Args:
        y_true: Actual labels
        y_pred: Predicted labels

    Returns:
        Classification report in sorted form.
    """
    base_report = M.classification_report(y_true,
                                          y_pred,
                                          output_dict=True,
                                          **kwargs)
    base_report_df = pd.DataFrame.from_dict(base_report).T
    class_wise_df = (base_report_df
                     .iloc[:-3]
                     .sort_values(by='f1-score'))
    summary_df = base_report_df.iloc[-3:]
    combined_df = pd.concat([class_wise_df, summary_df])
    combined_df['support'] = combined_df['support'].astype(int)
    return combined_df



from typing import List

from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import LogisticRegression, SGDClassifier
from sklearn.model_selection import ParameterGrid, GridSearchCV
from sklearn.multiclass import OneVsRestClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import Pipeline
from sklearn.tree import DecisionTreeClassifier

from fns.model_selection import view_result_table
import numpy as np
import os

__all__ = ['MultiLabelTextClassifier']


def generate_grid(grid):
    for grid_class, param_grid in grid.items():
        for p in ParameterGrid(param_grid):
            if grid_class in [LogisticRegression, SGDClassifier]:
                yield OneVsRestClassifier(grid_class(**p))
            else:
                yield grid_class(**p)


grid = {
    'models': {KNeighborsClassifier: {'n_neighbors': [5]},
               LogisticRegression: {'class_weight': ['balanced'],
                                    'max_iter': [5000],
                                    'C': np.linspace(0.001, 1, 50),
                                    'fit_intercept': [True, False]},
               SGDClassifier: {
                   'loss': ['hinge', 'log'],
                   'penalty': ['l1', 'l2', 'elasticnet'],
                   'alpha': np.logspace(-4, 1, 5),
                   'class_weight': ['balanced'],
                   'max_iter': [5000]
               },
               DecisionTreeClassifier: {
                   'class_weight': ['balanced']
               },
               RandomForestClassifier: {
                   'class_weight': ['balanced']
               }
               },
    'vectorizer': {TfidfVectorizer: {
        'ngram_range': [(1, 1), (1, 2), (1, 3), (2, 2), (2, 3)],
        'stop_words': [None, 'english'],
        'analyzer': ['word'],
        'binary': [True, False],
        'max_df': [1.0, 0.75, 0.5, 0.25],
        'strip_accents': ['unicode', None],
    }}
}


class MultiLabelTextClassifier:
    def __init__(self,
                 models: List = None,
                 vectorizers: List = None,
                 vectorizers_default=generate_grid(grid['vectorizer']),
                 models_default=generate_grid(grid['models'])):
        self.models = [] if not models else models
        self.vectorizers = [] if not vectorizers else vectorizers
        self.params = {'vectorizer': [*vectorizers_default,
                                      *self.vectorizers],
                       'model': [*models_default,
                                 *self.models],
                       }

    def fit(self,
            x_train,
            y_train,
            cv=3,
            scoring='f1_samples',
            memory=('/dev/shm/joblib' if os.name == 'posix' else None)):
        pipeline = Pipeline([('vectorizer', TfidfVectorizer()),
                             ('model', OneVsRestClassifier(LogisticRegression()))],
                            memory=memory)
        hpo = GridSearchCV(pipeline,
                           self.params,
                           cv=cv,
                           scoring=scoring,
                           n_jobs=-1,
                           verbose=1,
                           refit=True)
        hpo.fit(x_train, y_train)
        print(f'Best F1-score: {hpo.best_score_}')
        print(f'Best Estimator: {hpo.best_estimator_}')
        df = view_result_table(hpo)
        df.to_csv('grid-search.csv', index=False)
        return df
